use cgmath::{Matrix2, Rad, Vector2, Vector3, Zero};

use cgmath::Matrix4;
use serde::{Deserialize, Serialize};

use crate::prelude::*;
use anyhow::Result;

use super::angle;

pub struct Location {
    scale: Vector3<f32>,
    offset: Vector3<f32>,
    position: Vector3<f32>,
    rotation: Rad<f32>,

    movement_min_threshold: f32,

    transformation: Matrix4<f32>,

    draw: Option<UnsafeComponentStore<Draw>>,
    audio: Option<UnsafeComponentStore<Audio>>,
}

impl Clone for Location {
    fn clone(&self) -> Self {
        Self {
            scale: self.scale,
            offset: self.offset,
            position: self.position,
            rotation: self.rotation,

            movement_min_threshold: 0.4,

            transformation: self.transformation,

            draw: None,
            audio: None,
        }
    }
}

impl Location {
    pub const DEFAULT_DIRECTION: cgmath::Vector2<f32> = cgmath::Vector2 { x: 0.0, y: -1.0 };

    pub fn new_and_setup(entity: &mut EntityObject) -> Result<&mut Self> {
        let location = Self {
            scale: Vector3::new(1.0, 1.0, 1.0),
            offset: Vector3::zero(),
            position: Vector3::zero(),
            rotation: Rad(0.0),

            movement_min_threshold: 0.4,

            transformation: Matrix4::identity(),

            draw: None,
            audio: None,
        };

        entity.insert_component(location);
        Self::setup(entity.multi_mut())?;

        Ok(entity.get_component_mut::<Self>()?)
    }

    pub fn setup(mut multi_mut: MultiMut<'_>) -> Result<()> {
        let location = multi_mut.get::<Location>()?;
        let draw = multi_mut.get::<Draw>().ok();
        let audio = multi_mut.get::<Audio>().ok();

        if let Some(draw) = draw {
            location.set_draw_callback(draw);
            location.draw = Some(UnsafeComponentStore::from(draw));
        };

        location.audio = audio.map(UnsafeComponentStore::from);
        location.update_buffer()?;

        Ok(())
    }

    fn set_draw_callback(&self, draw: &mut Draw) {
        let location = UnsafeComponentStore::from(self);

        draw.set_model_buffer_callback(move |mesh| {
            mesh.model_buffer()
                .fill(&[(unsafe { location.as_mut() }.transformation_matrix()
                    * mesh.model_matrix())
                .into()])
                .unwrap();
        });
    }

    pub fn position(&self) -> Vector3<f32> {
        self.position
    }

    pub fn update(&mut self, position: Vector3<f32>, direction: Vector2<f32>) {
        if direction != Vector2::zero() {
            self.set_position(position);
            self.set_rotation(angle(direction));
        }
    }

    pub fn update_raw(&mut self, position: Vector3<f32>, angle: impl Into<Rad<f32>>) {
        self.set_position(position);
        self.set_rotation(angle.into());
    }

    pub fn set_scale_uniform(&mut self, scale: f32) {
        self.set_scale(cgmath::vec3(scale, scale, scale));
    }

    pub fn set_scale(&mut self, scale: Vector3<f32>) {
        if self.scale != scale {
            self.scale = scale;

            self.update_apply().unwrap();
        }
    }

    pub fn set_direction(&mut self, direction: Vector2<f32>) {
        if direction.magnitude() >= self.movement_min_threshold {
            self.set_rotation(angle(direction));
        }
    }

    pub fn set_rotation(&mut self, rotation: impl Into<Rad<f32>>) {
        let rotation = rotation.into();

        if self.rotation != rotation {
            self.rotation = rotation;

            self.update_apply().unwrap();
        }
    }

    pub fn set_position(&mut self, position: Vector3<f32>) {
        if self.position != position {
            self.position = position;

            self.update_apply().unwrap();
        }
    }

    pub fn set_offset(&mut self, offset: Vector3<f32>) {
        if self.offset != offset {
            self.offset = offset;

            self.update_apply().unwrap();
        }
    }

    pub fn rotation(&self) -> Rad<f32> {
        self.rotation
    }

    pub fn scale(&self) -> Vector3<f32> {
        self.scale
    }

    pub fn direction(&self) -> Vector2<f32> {
        Matrix2::from_angle(self.rotation) * Self::DEFAULT_DIRECTION
    }

    pub fn offset(&self) -> Vector3<f32> {
        self.offset
    }

    fn update_matrix(&mut self) {
        self.transformation = Matrix4::from_translation(self.position + self.offset)
            * Matrix4::from_angle_z(self.rotation)
            * Matrix4::from_nonuniform_scale(self.scale.x, self.scale.y, self.scale.z);
    }

    pub fn transformation_matrix(&self) -> &Matrix4<f32> {
        &self.transformation
    }

    fn update_apply(&mut self) -> Result<()> {
        self.update_matrix();

        self.update_buffer()?;

        Ok(())
    }

    pub(crate) fn update_buffer(&self) -> Result<()> {
        self.audio
            .as_ref()
            .map(|audio| unsafe { audio.as_mut() }.set_position(self.position + self.offset))
            .transpose()?;

        if let Some(draw) = &self.draw {
            assert!(draw.is_init(), "Run Location::setup first!");
            unsafe {
                self.apply_model_matrices(self.transformation_matrix(), draw.as_mut())?;
            }
        }

        Ok(())
    }

    /// # Safety
    ///
    /// Use only when you know what you are doing!

    pub unsafe fn apply_model_matrices(
        &self,
        transformation: &Matrix4<f32>,
        draw: &Draw,
    ) -> Result<()> {
        for mesh in draw.iter() {
            mesh.model_buffer()
                .fill(&[(transformation * mesh.model_matrix()).into()])?;
        }

        Ok(())
    }
}

impl EntityComponent for Location {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for Location {
    fn debug_name() -> &'static str {
        "Location"
    }
}

#[derive(Serialize, Deserialize)]
pub struct LocationCreateInfo {
    scale: Vector3<f32>,
    offset: Vector3<f32>,
    position: Vector3<f32>,
    rotation: Rad<f32>,
    movement_min_threshold: f32,
}

impl<'a> ComponentCreateInfo<'a> for LocationCreateInfo {}
