use crate::prelude::*;
use utilities::prelude::cgmath::{
    EuclideanSpace, InnerSpace, Matrix4, One, Point3, Rad, Vector3, Zero, vec3,
};

#[derive(Clone)]
pub struct Camera {
    eye_offset: Vector3<f32>,
    eye_dir: Vector3<f32>,
    center: Vector3<f32>,
    center_offset: Vector3<f32>,
    up_vector: Vector3<f32>,
    fov: f32,
    tan: f32,

    eye_rotation: f32,

    view: Matrix4<f32>,
    look_at: bool,

    update: bool,
}

impl Camera {
    pub fn new(fov: f32) -> Camera {
        let mut me = Camera {
            eye_offset: Vector3::zero(),
            eye_dir: Vector3::unit_y(),
            center: Vector3::zero(),
            center_offset: Vector3::zero(),
            up_vector: vec3(0.0, 0.0, 1.0),
            fov: 0.0,
            tan: 0.0,

            eye_rotation: 0.0,

            view: Matrix4::one(),
            look_at: true,

            update: true,
        };

        me.set_fov(fov);

        me
    }

    pub fn look_at(&mut self, look_at: bool) {
        self.look_at = look_at;
        self.update = true;
    }

    pub fn set_fov(&mut self, fov: f32) {
        let fovy = fov.to_radians() / 2.0;

        self.fov = fovy;
        self.tan = fovy.tan();
    }

    // basic setter & getter
    pub fn set_center(&mut self, center: Vector3<f32>) -> bool {
        if self.center != center {
            self.center = center;
            self.update = true;

            true
        } else {
            false
        }
    }

    pub fn center(&self) -> Vector3<f32> {
        self.center + self.center_offset
    }

    pub fn set_center_offset(&mut self, center_offset: Vector3<f32>) -> bool {
        if self.center_offset != center_offset {
            self.center_offset = center_offset;
            self.update = true;

            true
        } else {
            false
        }
    }

    pub fn set_rotation(&mut self, rotation: f32) {
        self.eye_rotation = rotation;
        self.update = true;
    }

    pub fn rotation(&self) -> f32 {
        self.eye_rotation
    }

    pub fn set_eye_offset(&mut self, eye_offset: Vector3<f32>) {
        self.eye_offset = eye_offset;
        self.update = true;
    }

    pub fn eye_offset(&self) -> Vector3<f32> {
        self.eye_offset
    }

    // view matrix handling
    fn eye_position(&self) -> Vector3<f32> {
        rotate_z(self.eye_offset, self.eye_rotation) + self.center + self.center_offset
    }

    pub fn set_eye_dir(&mut self, eye_dir: Vector3<f32>) {
        self.eye_dir = eye_dir;
        self.update = true;
    }

    pub fn view_matrix(&mut self) -> Matrix4<f32> {
        if self.update {
            self.calculate_view_matrix();
        }

        self.view
    }

    pub fn old_view_matrix(&self) -> Matrix4<f32> {
        self.view
    }

    pub fn calculate_view_matrix(&mut self) {
        self.update = false;
        self.view = if self.look_at {
            Matrix4::look_at_rh(
                Point3::from_vec(self.eye_position()),
                Point3::from_vec(self.center + self.center_offset),
                self.up_vector,
            )
        } else {
            Matrix4::look_to_rh(
                Point3::from_vec(self.center + self.center_offset),
                self.eye_dir,
                self.up_vector,
            )
        };
    }

    pub fn tan(&self) -> f32 {
        self.tan
    }

    pub fn fov(&self) -> Rad<f32> {
        Rad(self.fov)
    }

    /// returns (`view_direction`, `up`, `right`)
    pub fn axises(&self) -> (Vector3<f32>, Vector3<f32>, Vector3<f32>) {
        let view = (self.center() - self.eye_position()).normalize();
        let right = view.cross(self.up_vector).normalize();
        let up = right.cross(view);

        (view, up, right)
    }

    pub fn up(&self) -> Vector3<f32> {
        self.up_vector
    }

    pub fn set_up(&mut self, up: Vector3<f32>) {
        self.up_vector = up;
        self.update = true;
    }

    // calculation helper
    pub fn add_rotation(&mut self, rotation: f32) {
        self.eye_rotation += rotation;
        self.update = true;
    }

    pub fn move_on_plane(&mut self, distance: Vector3<f32>) {
        self.center += distance;
        self.update = true;
    }

    pub fn add_eye_offset(&mut self, delta: Vector3<f32>) {
        self.eye_offset += delta;
        self.update = true;
    }

    pub fn flip_view_on_z(&mut self, z: f32) {
        let old_z = self.eye_offset.z;
        let distance = -2.0 * (old_z - z);
        self.eye_offset.z = old_z + distance;
        self.update = true;
    }
}
