use std::sync::Arc;

use crate::prelude::*;

use anyhow::Result;
use utilities::{
    impl_reprc,
    prelude::cgmath::{InnerSpace, Rad, Vector3, Zero, vec3},
};

use utilities::prelude::cgmath::{Matrix4, SquareMatrix};

impl_reprc!(
    #[derive(Copy, Debug)]
    pub struct DirectionalLightData {
        #[assume_reprc]
        pub direction: Vector3<f32>,
        #[assume_reprc]
        pub t_min: f32,
        #[assume_reprc]
        pub color: Vector3<f32>,
        #[assume_reprc]
        pub power: f32,
    }
);

impl From<&Light> for DirectionalLightData {
    fn from(light: &Light) -> Self {
        DirectionalLightData {
            direction: light.direction(),
            t_min: light.t_min(),
            color: light.color(),
            power: light.power(),
        }
    }
}

impl_reprc!(
    #[derive(Copy, Debug)]
    pub struct PointLightData {
        #[assume_reprc]
        pub position: Vector3<f32>,
        #[assume_reprc]
        pub t_min: f32,
        #[assume_reprc]
        pub color: Vector3<f32>,
        #[assume_reprc]
        pub power: f32,
    }
);

impl From<&Light> for PointLightData {
    fn from(light: &Light) -> Self {
        PointLightData {
            position: light.position(),
            t_min: light.t_min(),
            color: light.color(),
            power: light.power(),
        }
    }
}

impl_reprc!(
    #[derive(Copy, Debug)]
    pub struct SpotLightData {
        #[assume_reprc]
        pub position: Vector3<f32>,
        #[assume_reprc]
        pub t_min: f32,
        #[assume_reprc]
        pub direction: Vector3<f32>,
        #[assume_reprc]
        pub angle: f32,
        #[assume_reprc]
        pub color: Vector3<f32>,
        #[assume_reprc]
        pub power: f32,
    }
);

impl From<&Light> for SpotLightData {
    fn from(light: &Light) -> Self {
        SpotLightData {
            position: light.position(),
            t_min: light.t_min(),
            direction: light.direction(),
            angle: light.angle().0,
            color: light.color(),
            power: light.power(),
        }
    }
}

#[derive(Debug, Clone)]
pub enum LightParameter {
    Position(Vector3<f32>),
    Direction(Vector3<f32>),
    Angle(Rad<f32>),
    Color(Vector3<f32>),
    Power(f32),
    TMin(f32),
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum LightType {
    Point,
    Direction,
    Spot,
}

#[derive(Clone)]
pub struct Light {
    #[allow(unused)]
    light_type: LightType,

    position: Vector3<f32>,
    direction: Vector3<f32>,
    angle: Rad<f32>,

    color: Vector3<f32>,
    power: f32,
    t_min: f32,

    view: View,

    update_callback: Option<Arc<dyn Fn(LightParameter, &mut View) -> Result<()> + Send + Sync>>,
}

impl Light {
    pub fn point_light(device: &Arc<Device>) -> Result<Light> {
        Self::new(LightType::Point, device)
    }

    pub fn directional_light(device: &Arc<Device>) -> Result<Light> {
        Self::new(LightType::Direction, device)
    }

    pub fn spot_light(device: &Arc<Device>) -> Result<Light> {
        Self::new(LightType::Spot, device)
    }

    #[inline]
    fn new(light_type: LightType, device: &Arc<Device>) -> Result<Light> {
        Ok(Light {
            #[allow(unused)]
            light_type,

            position: Vector3::zero(),
            direction: vec3(0.0, 0.0, -1.0),
            angle: Rad(180.0),

            color: vec3(1.0, 1.0, 1.0),
            power: 60.0,

            t_min: 0.0,

            update_callback: None,

            view: View::new(device, Matrix4::identity(), 45.0, false)?,
        })
    }

    pub fn set_position(&mut self, new_position: Vector3<f32>) -> Result<()> {
        if new_position != self.position {
            self.position = new_position;

            if let Some(callback) = &self.update_callback {
                callback(LightParameter::Position(new_position), &mut self.view)?;
            }
        }

        Ok(())
    }

    pub fn set_direction(&mut self, new_direction: Vector3<f32>) -> Result<()> {
        let normalized = new_direction.normalize();

        if normalized != self.direction {
            self.direction = normalized;

            if let Some(callback) = &self.update_callback {
                callback(LightParameter::Direction(normalized), &mut self.view)?;
            }
        }

        Ok(())
    }

    pub fn set_color(&mut self, color: Vector3<f32>) -> Result<()> {
        if color != self.color {
            self.color = color;

            if let Some(callback) = &self.update_callback {
                callback(LightParameter::Color(self.color), &mut self.view)?;
            }
        }

        Ok(())
    }

    pub fn set_power(&mut self, power: f32) -> Result<()> {
        if power != self.power {
            self.power = power;

            if let Some(callback) = &self.update_callback {
                callback(LightParameter::Power(self.power), &mut self.view)?;
            }
        }

        Ok(())
    }

    pub fn set_angle(&mut self, new_angle: impl Into<Rad<f32>>) -> Result<()> {
        let casted_angle = new_angle.into();

        if casted_angle != self.angle {
            self.angle = casted_angle;

            if let Some(callback) = &self.update_callback {
                callback(LightParameter::Angle(casted_angle), &mut self.view)?;
            }
        }

        Ok(())
    }

    pub fn set_t_min(&mut self, t_min: f32) -> Result<()> {
        if t_min != self.t_min {
            self.t_min = t_min;

            if let Some(callback) = &self.update_callback {
                callback(LightParameter::TMin(t_min), &mut self.view)?;
            }
        }

        Ok(())
    }

    pub fn position(&self) -> Vector3<f32> {
        self.position
    }

    pub fn direction(&self) -> Vector3<f32> {
        self.direction
    }

    pub fn color(&self) -> Vector3<f32> {
        self.color
    }

    pub fn power(&self) -> f32 {
        self.power
    }

    pub fn angle(&self) -> Rad<f32> {
        self.angle
    }

    pub fn t_min(&self) -> f32 {
        self.t_min
    }

    #[allow(unused)]

    pub(crate) fn set_update_callback<F>(&mut self, f: F)
    where
        F: Fn(LightParameter, &mut View) -> Result<()> + Send + Sync + 'static,
    {
        self.update_callback = Some(Arc::new(f));
    }

    #[allow(unused)]

    pub(crate) fn clear_update_callback(&mut self) {
        self.update_callback = None;
    }

    #[allow(unused)]
    pub(crate) fn light_type(&self) -> LightType {
        self.light_type
    }

    pub fn eq(&self, other: &Light) -> bool {
        self.light_type == other.light_type
            && self.position == other.position
            && self.direction == other.direction
            && self.angle == other.angle
            && self.color == other.color
            && self.power == other.power
            && self.t_min == other.t_min
    }
}
