use crate::prelude::*;

use anyhow::Result;
use asset::RaytracerDesciptorLayouts as AssetLayouts;
use utilities::{
    impl_reprc,
    prelude::cgmath::{Vector2, Vector3},
};

use super::{
    super::{
        shared::{bufferhandler::*, safecommandbuffer::SafeCommandBuffer},
        RaytracerDesciptorLayouts,
    },
    raytraceshader::RayTracingShader,
};

use std::sync::{
    atomic::{AtomicU32, Ordering::SeqCst},
    Arc, Mutex,
};

macro_rules! update_iter {
    ( $buffer_handler: ident, $count: ident, $object_id: ident, $(($name: ident, $primitive_type: ty),)+ ) => {
        paste::expr! {
            $(
                let [<$name _iter>] = {
                    let count_clone = $count.clone();
                    let object_id_clone = $object_id.clone();

                    Box::new(move |
                        renderer: &SafeCommandBuffer<$primitive_type>,
                        recorder: &mut CommandBufferRecorder<'_>,
                        primitive: &PrimitiveData<$primitive_type>,
                        additional_sets: &[&Arc<DescriptorSet>],
                        _view: &View
                    | {
                            let vertex_count = primitive.vertex_buffer.size() as u32;

                            renderer.dispatch(
                                recorder,
                                &[&primitive.descriptor_set, additional_sets[0]],
                                Some((count_clone.load(SeqCst), object_id_clone.load(SeqCst))),
                                vertex_count
                            )?;

                            count_clone.fetch_add(vertex_count, SeqCst);
                            object_id_clone.fetch_add(1, SeqCst);

                            Ok(())
                        }
                    )
                };

                $buffer_handler.[<set_ $name _function>]([<$name _iter>]);
            )+
        }
    };
}

// just a define, this type is output in 'animation.comp' shader and input in 'closesthit.rchit'
impl_reprc!(
    #[derive(Copy)]
    pub struct VertexData {
        #[assume_reprc]
        position: Vector3<f32>,
        #[assume_reprc]
        normal: Vector3<f32>,

        // in case of colored vertex uv (x, y) and tangent (x, y, z)
        // will be identifier (i == -1.0) an color(r, g, b, a)
        #[assume_reprc]
        uv: Vector2<f32>,
        #[assume_reprc]
        tangent: Vector3<f32>,
        #[assume_reprc]
        material_index: u32,
    }
);

pub struct ComputeBasedAnimator {
    device: Arc<Device>,

    count: Arc<AtomicU32>,
    object_id: Arc<AtomicU32>,
    buffer_handler: BufferHandler,

    scene_buffer: Arc<Buffer<VertexData>>,
    descriptor_set: Arc<DescriptorSet>,
}

impl ComputeBasedAnimator {
    pub fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
    ) -> Result<ComputeBasedAnimator> {
        let (buffer_handler, count, object_id) =
            Self::init_buffer_handler(device, queue, image_count)?;

        let layout = RaytracerDesciptorLayouts::vertex_type_descriptor_layout(device.clone())?;

        let pool = DescriptorPool::builder()
            .set_layout(layout)
            .build(device.clone())?;

        let descriptor_set = DescriptorPool::prepare_set(&pool).allocate()?;

        Ok(ComputeBasedAnimator {
            device: device.clone(),

            count,
            object_id,
            buffer_handler,

            scene_buffer: Self::create_scene_buffer(device, 1, &descriptor_set)?,
            descriptor_set,
        })
    }

    pub fn buffer(&self) -> &Arc<Buffer<VertexData>> {
        &self.scene_buffer
    }

    pub fn invalidate(&mut self) {
        self.buffer_handler.invalidate();
    }

    pub fn animate(
        &mut self,
        content: &[&EntityObject],
        vertex_count: VkDeviceSize,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        index: usize,
        view: &View,
    ) -> Result<Option<&Arc<Buffer<VertexData>>>> {
        // reset count
        self.count.store(0, SeqCst);

        // reset object id counter
        self.object_id.store(0, SeqCst);

        let mut opt_buffer = None;

        if self.scene_buffer.size() != vertex_count {
            self.scene_buffer =
                Self::create_scene_buffer(&self.device, vertex_count, &self.descriptor_set)?;
            opt_buffer = Some(&self.scene_buffer);
        }

        let mut buffers = Vec::new();

        self.buffer_handler.execute(
            None,
            content,
            &[&self.descriptor_set],
            &mut buffers,
            index,
            view,
        )?;

        buffer_recorder.buffer_barrier(
            &self.scene_buffer,
            0,
            VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR,
            VK_ACCESS_SHADER_WRITE_BIT,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
        );

        buffer_recorder.execute_commands(&buffers);

        buffer_recorder.buffer_barrier(
            &self.scene_buffer,
            VK_ACCESS_SHADER_WRITE_BIT,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
            VK_ACCESS_SHADER_READ_BIT,
            VK_PIPELINE_STAGE_RAY_TRACING_SHADER_BIT_KHR,
        );

        Ok(opt_buffer)
    }

    fn create_scene_buffer(
        device: &Arc<Device>,
        size: VkDeviceSize,
        set: &Arc<DescriptorSet>,
    ) -> Result<Arc<Buffer<VertexData>>> {
        let buffer = Buffer::builder()
            .set_usage(
                VK_BUFFER_USAGE_STORAGE_BUFFER_BIT
                    | VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT
                    | VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR,
            )
            .set_memory_usage(MemoryUsage::GpuOnly)
            .set_size(size)
            .build(device.clone())?;

        set.update(&[DescriptorWrite::storage_buffers(0, &[&buffer])])?;

        Ok(buffer)
    }

    #[inline]
    fn init_buffer_handler(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
    ) -> Result<(BufferHandler, Arc<AtomicU32>, Arc<AtomicU32>)> {
        let input_model_layout = AssetLayouts::descriptor_layout(device.clone())?;
        let input_model_bone_layout = AssetLayouts::bone_descriptor_layout(device.clone())?;
        let output_buffer_layout =
            RaytracerDesciptorLayouts::vertex_type_descriptor_layout(device.clone())?;

        let shader = RayTracingShader::animation(device)?;
        let push_constant = [VkPushConstantRange::new(VK_SHADER_STAGE_COMPUTE_BIT, 0, 8)];

        let create_info = BufferHandlerCreateInfo::new(
            &shader,
            device,
            queue,
            image_count,
            BufferHandlerType::compute(
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_bone_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_bone_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
                PipelineLayoutInfo {
                    descriptor_layouts: vec![
                        input_model_bone_layout.as_ref(),
                        output_buffer_layout.as_ref(),
                    ],
                    push_constants: &push_constant,
                },
            ),
        );

        let mut buffer_handler = BufferHandler::new(&create_info, false)?;

        let count = Arc::new(AtomicU32::new(0));
        let object_id = Arc::new(AtomicU32::new(0));

        update_iter!(
            buffer_handler,
            count,
            object_id,
            (position, PositionOnly),
            (position_coloruv, PositionColorUV),
            (position_normal, PositionNormal),
            (position_normal_color, PositionNormalColor),
            (position_normal_coloruv, PositionNormalColorUV),
            (position_normal_coloruv_bone, PositionNormalColorUVBone),
            (
                position_normal_coloruv_normaluv,
                PositionNormalColorUVNormalUV
            ),
            (
                position_normal_coloruv_normaluv_bone,
                PositionNormalColorUVNormalUVBone
            ),
        );

        Ok((buffer_handler, count, object_id))
    }
}
