use quote::{format_ident, ToTokens};
use syn::{Expr, PathArguments, Type};

#[derive(Debug)]
pub struct Array {
    pub full_type: proc_macro2::TokenStream,
    pub inner_type: proc_macro2::TokenStream,
    pub length: proc_macro2::TokenStream,
}

#[derive(Debug)]
pub struct Vector {
    pub full_type: proc_macro2::TokenStream,
    pub inner_type: proc_macro2::TokenStream,
}

#[derive(Debug)]
pub enum DataType {
    Array(Array),
    Vector(Vector),
    Other(proc_macro2::TokenStream),
}

impl DataType {
    pub fn new(field_type: &Type) -> Self {
        match field_type {
            Type::Array(array) => Self::Array(Array {
                full_type: array.to_token_stream(),
                inner_type: array.elem.to_token_stream(),
                length: Self::get_array_len(&array.len),
            }),

            Type::BareFn(_) => {
                todo!("Field Type Bare Fn")
            }
            Type::Group(_) => {
                todo!("Field Type Group")
            }
            Type::ImplTrait(_) => {
                todo!("Field Type Impl Trait")
            }
            Type::Infer(_) => {
                todo!("Field Type Infer")
            }
            Type::Macro(_) => {
                todo!("Field Type Macro")
            }
            Type::Never(_) => {
                todo!("Field Type Never")
            }
            Type::Paren(_) => {
                todo!("Field Type Paren")
            }
            Type::Path(path) => {
                let segment = &path.path.segments[0];

                if segment.ident == format_ident!("Vec") {
                    let inner_type = match &segment.arguments {
                        PathArguments::AngleBracketed(args) => args.args.to_token_stream(),
                        _ => unreachable!(),
                    };

                    return Self::Vector(Vector {
                        full_type: path.path.to_token_stream(),
                        inner_type,
                    });
                }

                Self::Other(path.path.to_token_stream())
            }
            Type::Ptr(_) => {
                todo!("Field Type Pointer")
            }
            Type::Reference(_) => {
                todo!("Field Type Reference")
            }
            Type::Slice(_) => {
                todo!("Field Type Slice")
            }
            Type::TraitObject(_) => {
                todo!("Field Type Trait Object")
            }
            Type::Tuple(_) => {
                todo!("Field Type Tuple")
            }
            Type::Verbatim(_) => {
                todo!("Field Type Verbatim")
            }
            &_ => {
                todo!("exhaustive")
            }
        }
    }

    fn get_array_len(e: &Expr) -> proc_macro2::TokenStream {
        match e {
            Expr::Path(path) => path.path.to_token_stream(),
            Expr::Lit(lit) => lit.lit.to_token_stream(),

            _ => todo!(),
        }
    }
}

impl ToTokens for DataType {
    fn to_tokens(&self, tokens: &mut proc_macro2::TokenStream) {
        let token_stream = match self {
            Self::Array(array) => &array.full_type,
            Self::Vector(vector) => &vector.full_type,
            Self::Other(ts) => ts,
        };

        token_stream.to_tokens(tokens);
    }
}
